@@ -191,9 +191,31 @@ sub _build_formats {
   };
 }
 
+sub _bundle_ref {
+  my ($self, $state, $source, $target) = @_;
+
+  # Recurse into the inner $ref to avoid invalid schema
+  my $source_state = $self->store->resolve($source->{'$ref'}, $state);
+  return $self->_bundle_ref($source_state, $source_state->{schema}, $target) if $source_state->{schema}{'$ref'};
+
+  # Replace "paths" inline
+  if (@{$state->{schema_path}} == 2 and $state->{schema_path}[0] eq 'paths') {
+    delete $state->{seen_schema}{$target};    # path definitions should not recurse
+    return [$source_state, $source_state->{schema}, $target, $state->{schema_path}];
+  }
+
+  return $self->SUPER::_bundle_ref($state, $source, $target);
+}
+
 sub _bundle_ref_path_expand {
-  my ($self, $ref) = @_;
-  return $ref =~ m!\b(definitions|parameters|responses)/(.+)$! ? ($1, $2) : ('definitions', $ref);
+  my ($self, $state, $ref) = @_;
+  return parameters  => $ref if $state->{schema}{in};
+  return definitions => $ref if $state->{schema}{properties};
+  return ($1, $2 ? ($2) : ()) if $ref =~ m!\b(x-[^/]+)/?\b(.*)!;
+  return (definitions => $ref) unless $ref =~ m!\b(definitions|parameters|responses)\b(.+)$!;
+  my ($section, $path) = ($1, $2);
+  $path =~ s!^\.?\b(json|yaml)\W*!!;
+  return ($section => $path);
 }
 
 sub _coerce_arrays {
